/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.dev.transport;

import java.io.File;
import java.io.FilenameFilter;
import java.text.DateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import org.tmatesoft.svn.core.ISVNDirEntryHandler;
import org.tmatesoft.svn.core.ISVNLogEntryHandler;
import org.tmatesoft.svn.core.SVNDepth;
import org.tmatesoft.svn.core.SVNDirEntry;
import org.tmatesoft.svn.core.SVNException;
import org.tmatesoft.svn.core.SVNLogEntry;
import org.tmatesoft.svn.core.SVNLogEntryPath;
import org.tmatesoft.svn.core.SVNURL;
import org.tmatesoft.svn.core.io.SVNRepository;
import org.tmatesoft.svn.core.wc.ISVNOptions;
import org.tmatesoft.svn.core.wc.SVNRevision;
import org.tmatesoft.svn.core.wc.SVNUpdateClient;
import org.tmatesoft.svn.core.wc.SVNWCClient;
import org.tmatesoft.svn.core.wc.SVNWCUtil;

import com.zfabrik.util.fs.FileUtils;

public class Transport {

    public static Collection<ProjectDelta> computeProjectDeltas(
    		final Collection<String> tps, 
    		final SVNRepository repo, 
    		SVNRevision startRev, 
    		SVNRevision endRev, 
    		final String user, 
    		final Logger logger
    ) {
    	return computeProjectDeltas(tps,repo,startRev,endRev,user,logger,false);
    }	
    
    /**
     * Traverse log history in the source side and identify what projects have
     * been added, modified, or deleted.
     *   
     * @return
     */
    public static Collection<ProjectDelta> computeProjectDeltas(
    		final Collection<String> tps, 
    		final SVNRepository repo, 
    		SVNRevision startRev, 
    		SVNRevision endRev, 
    		final String user, 
    		final Logger logger,
    		final boolean flat
    ) {
        final Map<String,ProjectDelta> l = new HashMap<String,ProjectDelta>();
        try {
        	// rp is where we start looking in the repo
            final String rp = repo.getRepositoryPath("");
            if (endRev.getNumber()<0) {
                endRev = SVNRevision.create(repo.getLatestRevision());
            }
            if (startRev.getNumber()<0) { 
                throw new IllegalArgumentException("Start revision must be a number (unlike HEAD et al)");
            }
            
            // pl is where the remaining path, i.e. after the repository URL starts.
            final int pl = rp.length()+1;
            repo.log(new String[]{rp},
                    startRev.getNumber()+1,
                    endRev.getNumber(),
                    true,
                    true,
                    new ISVNLogEntryHandler(){

                SVNLogEntryPath lp;
                String p;
                ProjectDelta pd;

                public void handleLogEntry(SVNLogEntry le) throws SVNException {

                    // if a user filter is set and the log entry is not from the current user, ignore it
                    if (user != null && ! user.equals(le.getAuthor())) {
                        return;
                    }

                    boolean exactly;
                    for (Map.Entry<String,SVNLogEntryPath> e : le.getChangedPaths().entrySet()) {
                        lp = (SVNLogEntryPath) e.getValue();
                        p = (String) e.getKey();

                        if ((p.length()>=pl) && (p.startsWith(rp))) {
                        	String pa = p.substring(pl);
                            // parse path pa which is <module>/<module files>
                        	if (flat) {
                        		// in that case, no real modules, just the empty "" module
                        		exactly = pa.length()==0;
                        		pa = "";
                        	} else {
	                            int pe = pa.indexOf('/');
	                            exactly = (pe<0);
	                            if (!exactly) {
	                                pa = pa.substring(0,pe);
	                            }
                        	}
                            // pa is the module now.
                            pd = l.get(pa);
                            if (pd==null) {
                                pd = new ProjectDelta(pa);
                                l.put(pa,pd);
                            }
                            // add the rev to the ones having impact on this project
                            pd.getRevs().add(le.getRevision());
                            // determine the overall meaning for the target 
                            switch (lp.getType()) {
                            case SVNLogEntryPath.TYPE_DELETED:
                                if (tps.contains(pd.getName())) {
                                    if (exactly) {
                                        pd.setType(ProjectDelta.DEL);
                                    } else {
                                        pd.setType(ProjectDelta.MOD);
                                    }
                                } else {
                                    if (!exactly) {
                                        pd.setType(ProjectDelta.MOD);
                                    } else {
                                        // remove from projects, as it does not exist on the target and has been deleted here
                                        l.remove(pa);
                                    }
                                }
                                break;
                            case SVNLogEntryPath.TYPE_ADDED:
                                if (exactly) {
                                    if (tps.contains(pd.getName())) {
                                        // change into modified if existing on target side
                                        pd.setType(ProjectDelta.MOD);
                                    } else {
                                        pd.setType(ProjectDelta.ADD);
                                    }
                                } else {
                                    if (pd.getType()!=ProjectDelta.ADD) {
                                        // unless it is an add, it becomes a MOD 
                                        pd.setType(ProjectDelta.MOD);
                                    }
                                }
                                break;
                            case SVNLogEntryPath.TYPE_MODIFIED:
                            case SVNLogEntryPath.TYPE_REPLACED:
                                if (pd.getType()!=ProjectDelta.ADD) {
                                    // unless it is an add, it is a MOD 
                                    pd.setType(ProjectDelta.MOD);
                                }
                                break;
                            }
                        }
                    }
                }
            });
        } catch (SVNException e) {
            throw new RuntimeException("Failed to read repository history",e);
        }
        return l.values();
    }

    /**
     * Intersect those that have been modified and deleted with the projects
     * existing on the target side.
     */
    public static void removeNonApplicableProjects(Collection<String> tps, Collection<ProjectDelta> delta) {
        Set<ProjectDelta> tbr = new HashSet<ProjectDelta>();
        for (ProjectDelta pd : delta) {
            if ((pd.getType()!=ProjectDelta.ADD) && !tps.contains(pd.getName())) {
                tbr.add(pd);
            }
        }
        delta.removeAll(tbr);
    }

    /**
     * Removes projects that are not touched by the user 
     */
    public static void removeNonUserProjects(Collection<String> tps, Collection<ProjectDelta> delta) {
        Set<ProjectDelta> tbr = new HashSet<ProjectDelta>();
        for (ProjectDelta pd : delta) {
            if ((pd.getType()!=ProjectDelta.ADD) && !tps.contains(pd.getName())) {
                tbr.add(pd);
            }
        }
        delta.removeAll(tbr);
    }

    /**
     * fetch projects on target side
     * @param tr
     * @param trev
     * @return
     */
    public static Collection<String> getTargetProjects(SVNRepository repo, SVNRevision rev) {
        try {
            final Set<String> result = new HashSet<String>();
            repo.getDir("", rev.getNumber(),null, new ISVNDirEntryHandler() {
                public void handleDirEntry(SVNDirEntry de) throws SVNException {
                    result.add(de.getName());
                }
            });
            return result;
        } catch (SVNException e) {
            throw new RuntimeException("Failed to read target projects list",e);
        }
    }


    /**
     * Creates workspace that contains an application of the delta from 
     * a project level perspective. That is: 
     * <ul>
     * <li>project added/modified, target project exists: 
     * 	<ol>
     * 	<li>Check out target project</li>
     *  <li>copy all files from source project over. Files that do not exist on target side mark as added, 
     *  files that do not exist on the source side mark as deleted.
     *  </li>
     *  </ol></li>
     * <li>project added/modified, target project does not exist:
     * 	<ol>
     * <li>add source project files</li>
     * </ol></li>
     * <li>project deleted, target project exists:
     * <ol><li>delete target project</li></ol>
     * </li>
     * Since subversion assumes that lastmodified is changed when a file is changed (which is REALLY stupid)
     * we make sure that fast operations do not confuse subversion by creating a 1s delta.
     * {@link http://lists.alioth.debian.org/pipermail/pkg-subversion-maintainers/2006-June/000726.html}
     * </ul>
     * 
     * if srev is null, the source-revision number will be the maximum revision number of the ProjectDelta!
     */
    public static void createDeltaWorkSpace(
            SVNRepository source, 
            SVNRevision srev, 
            SVNRepository target, 
            SVNRevision trev, 
            Collection<ProjectDelta> delta, 
            File folder, 
            Logger logger, 
            ICancellation cancellation
    ) {
    	
        if (folder.exists()) {
            FileUtils.delete(folder);
        }
        folder.mkdirs();
        File temp = null;
        try {
            temp = File.createTempFile("z2f_", "tmp");
            temp.deleteOnExit();
            File td = temp.getParentFile();
            temp = new File(td, "z2__" + temp.getName().substring(4));
            temp.mkdir();
            if (logger!=null) logger.fine("Using temp folder "+temp);

            ISVNOptions op = SVNWCUtil.createDefaultOptions(false);
            SVNUpdateClient tuc = new SVNUpdateClient(target.getAuthenticationManager(),op);
            tuc.setUpdateLocksOnDemand(true);
            SVNUpdateClient suc = new SVNUpdateClient(source.getAuthenticationManager(),op);
            suc.setUpdateLocksOnDemand(true);
            SVNWCClient     wc = new SVNWCClient(target.getAuthenticationManager(),op);
            
            if (logger!=null) logger.info("Checking out "+target.getLocation()+"@"+trev+" to "+folder+" non-recursively");
            tuc.doCheckout(target.getLocation(),folder,trev, trev, SVNDepth.EMPTY,  false);

            for (ProjectDelta pd:delta) {
                SVNRevision sourceRev = (srev != null)? srev : SVNRevision.create(pd.getMaxRev());
                
                if (logger!=null) {
                    logger.info("Starting "+pd.getName());
                }
                if ((cancellation!=null) && (cancellation.isCancelled())) break;
                File tf = new File(folder,pd.getName());
                SVNURL su = source.getLocation().appendPath(pd.getName(), false);
                if (target.info(pd.getName(), -1)!=null) {
                    // target exists. Check out!
                    SVNURL tu = target.getLocation().appendPath(pd.getName(),false);
                    
                    int i = 10;
                    while (i>0) {
                        if (logger!=null) logger.info("Retrieving "+tu+"@"+trev+" to "+tf+ " ("+i+" attempts left)");
                        try {
                            tuc.doUpdate(tf,trev, (pd.getType()!=ProjectDelta.DEL? SVNDepth.INFINITY : SVNDepth.EMPTY), false, true);
                            i=0;
                        } catch (SVNException svne) {
                            if (svne.getMessage().contains("SSL peer shut down incorrectly")) {
                                logger.info("Timeout during update - will retry");
                                i--;
                            } else {
                                throw svne;
                            }
                        }
                    }

                    if (pd.getType()==ProjectDelta.DEL) {
                        if (logger!=null) logger.info("Scheduling "+pd.getName()+" for deletion");
                        wc.doDelete(tf, true,false);
                    } else {
                        // get all files from the source and write them to the target. Add/delete as necessary
                        File   sf = new File(temp,pd.getName());
                        sf.mkdir();
                        if ((cancellation!=null) && (cancellation.isCancelled())) break;
                        if (logger!=null) logger.info("Exporting "+su+"@"+sourceRev);
                        suc.doExport(su, sf, sourceRev,sourceRev,  null, true, SVNDepth.INFINITY);
                        if (logger!=null) logger.info("Structural merge...");
                        _merge(sf,tf,wc,logger);
                        FileUtils.delete(sf);
                    }
                } else {
                    if ((pd.getType()==ProjectDelta.ADD)  || (pd.getType()==ProjectDelta.MOD)) {
                        tf.mkdir();
                        if (logger!=null) logger.info("Exporting "+su+"@"+srev);
                        if ((cancellation!=null) && (cancellation.isCancelled())) break;
                        suc.doExport(su, tf, sourceRev,sourceRev,  null, true, SVNDepth.INFINITY);
                        if (logger!=null) logger.info("Scheduling "+pd.getName()+" for add");
                        if ((cancellation!=null) && (cancellation.isCancelled())) break;
                        wc.doAdd(tf, true, false, false, SVNDepth.INFINITY,false,true);
                    } 
                }
                if (logger!=null) {
                    logger.info("Completed project "+pd.getName());
                }
            }
            if (logger!=null) {
                logger.info("Workspace creation completed.");
            }
        } catch (Exception e) {
            throw new RuntimeException("Failed to create delta workspace at "+folder,e);
        } finally {
            if (temp!=null) {
                FileUtils.delete(temp);
            }
        }
        if ((cancellation!=null) && (cancellation.isCancelled())) {
            if (logger!=null) logger.info("Workspace creation cancelled.");
        }
    }

    
//    @SuppressWarnings("serial")
//    private static class BreakIterationException extends RuntimeException {}
    
//    private static class CountDown {
//        private int c;
//        CountDown(int start) {
//            this.c = start;
//        }
//        boolean dec() {
//            return --c <= 0;
//        }
//    }

    private static FilenameFilter NON_SVN = new FilenameFilter() {

        public boolean accept(File dir, String name) {
            return !".svn".equals(name);
        }

    };

    /*
     */
    private static void _merge(File sf, File tf, SVNWCClient wc, Logger logger) {
        try {
            long timestamp = System.currentTimeMillis()+1000;
            Set<String> sfiles = new HashSet<String>(Arrays.asList(sf.list(NON_SVN)));
            Set<String> tfiles = new HashSet<String>(Arrays.asList(tf.list(NON_SVN)));
            File s,t;
            boolean ex;
            for (String sfn:sfiles) {
                s = new File(sf,sfn);
                t = new File(tf,sfn);
                ex=tfiles.remove(sfn);
                if (s.isDirectory()) {
                    if (!ex) {
                        if (logger!=null) logger.fine("[A] "+t);
                        wc.doAdd(t, true, true, false, SVNDepth.INFINITY,false,true);
                    } else 
                        if (!t.isDirectory()) {
                            if (logger!=null) logger.fine("[D] "+t);
                            wc.doDelete(t, true, false);
                            if (logger!=null) logger.fine("[A] "+t);
                            wc.doAdd(t, true, true, false, SVNDepth.EMPTY,false,true);
                        }
                    _merge(s,t,wc,logger);
                } else {
                    boolean dir = t.isDirectory();
                    if (dir) {
                        if (logger!=null) logger.fine("[D] "+t);
                        wc.doDelete(t, true, false);
                    }
                    FileUtils.copy(s,t,null);
                    t.setLastModified(timestamp=Math.max(System.currentTimeMillis(),timestamp));
                    if (dir || (!ex)) {
                        if (logger!=null) logger.fine("[A] "+t);
                        wc.doAdd(t, true, false, false,SVNDepth.EMPTY,false,true);
                    } else {
                        if (logger!=null) logger.fine("[M] "+t);
                    }
                }
            }
            // delete remaining
            for (String tfn:tfiles) {
                t = new File(tf,tfn);
                wc.doDelete(t, true, false);
                if (logger!=null) logger.fine("[D] "+t);
            }
        } catch (Exception e) {
            throw new RuntimeException("Failed to merge structures at (source: "+sf+", target: "+tf+")",e);
        }
    }
    
    private final static Pattern JIRA_COMMANDS = Pattern.compile("(?:^|\\s)#(\\p{Alpha}+)");
    private final static Pattern REDMINE_LOG = Pattern.compile("(^|\\s)@(?:\\w+)");
    

    public static void writeSummary(Collection<ProjectDelta> delta, SVNRepository srepo, SVNRevision st, SVNRevision en, boolean complete, final Logger l) throws Exception {
    	writeSummary(delta, srepo, null, st, en, complete, l);
    }

    public static void writeSummary(Collection<ProjectDelta> delta, SVNRepository srepo, SVNRepository trepo, SVNRevision st, SVNRevision en, boolean complete, final Logger l) throws Exception
    {
        // collect all revs
        final SortedSet<Long> sd = new TreeSet<Long>();
        for (ProjectDelta pd:delta) { sd.addAll(pd.getRevs()); }
        if (sd.size()==0) {
            l.warning("No revisions found!");
        } else {

        	if (trepo!=null) {
                l.info(
                        trepo.getLocation()+
                        " from "
                    );
        	}
        	
            l.info(
                srepo.getLocation()+
                "@"+
                st+
                ":"+
                en+ 
                (!complete? " PARTIAL":"")
            );
            
            
            l.info("");
            
            // now fetch descriptions
            srepo.log(null, sd.last(), sd.first(), false, false, new ISVNLogEntryHandler() {

                DateFormat df = DateFormat.getDateTimeInstance(DateFormat.MEDIUM , DateFormat.MEDIUM);

                public void handleLogEntry(SVNLogEntry le) throws SVNException {
                    if (sd.contains(le.getRevision())) {
                        l.info("rev:"+le.getRevision()+", author:"+le.getAuthor()+", time:"+df.format(le.getDate()));
                        String m = le.getMessage();
                        String n;
                        int a=0,d=0,e;
                        while (d>=0) {
                            e=d=m.indexOf(10,a);
                            if ((e>0) && (m.charAt(e-1)==13)) e--;
                            n = (e<0? m.substring(a) : m.substring(a,e));
                            
                            // escape JIRA commands
                            n=JIRA_COMMANDS.matcher(n).replaceAll(" #_$1");
                            n=REDMINE_LOG.matcher(n).replaceAll(" @_$1");
                            
                            l.info("> "+n);
                            a=d+1;
                        }
                        l.info("");
                    }
                }

            });
        }

    }
    
    

}
