/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.tx;

import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.Status;
import javax.transaction.UserTransaction;

/**
 * This class provides utilities around transaction management with Z2's built-in JTA implementation.
 * 
 * @author hb
 *
 */
public class TransactionUtil {

	/**
	 * Use this method to execute work in a transaction. This is just a utility that helps implementing
	 * for example Transaction Servlet Filters that begin and end transactions at the request boundaries.
	 * 
	 * If there is an ongoing transaction already, this method will not begin a new one.
	 * 
	 * 
	 * @param ut the {@link UserTransaction} instance of the transaction management system in place
	 * @param callable The {@link Callable} implementation that defines the work.
	 * @return The result of the callable invocation.
	 * @throws Exception
	 */
	public static <V> V run(UserTransaction ut, Callable<V> callable) throws Exception {
		if (ut.getStatus()!=Status.STATUS_ACTIVE) {
			ut.begin();
			try {
				return callable.call();
			} catch (Exception e) {
				logger.log(Level.WARNING,"Error at transaction boundary. Setting transaction to rollbackonly",e);
				ut.setRollbackOnly();
				throw e;
			} finally {
				if (ut.getStatus()==Status.STATUS_ACTIVE) {
					ut.commit();
				} else {
					ut.rollback();
				}
			}
		} else {
			// there is a tx ongoing already. Simply work.
			return callable.call();
		}
	}
	
	private final static Logger logger = Logger.getLogger(TransactionUtil.class.getName());
}
